/* This code is based on thamlt by an anonymous hacker. */
#include <config.h>

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <errno.h>
#include <libgen.h>

#include "thcrypt.h"
#include "thlzss.h"

struct {
	unsigned char key;
	unsigned char step;
	unsigned int block;
	unsigned int limit;
} g_cryprm[] = {
	/* key  step  block  limit */
	{ 0x1b, 0x37, 0x40,  0x2800 },
	{ 0x51, 0xe9, 0x40,  0x3000 },
	{ 0xc1, 0x51, 0x80,  0x3200 },
	{ 0x03, 0x19, 0x400, 0x7800 },
	{ 0xab, 0xcd, 0x200, 0x2800 },
	{ 0x12, 0x34, 0x80,  0x3200 },
	{ 0x35, 0x97, 0x80,  0x2800 },
	{ 0x99, 0x37, 0x400, 0x2000 }
};

static unsigned int
th_get_cryprm_index(const char *name)
{
	char index = 0;
	while(*name) index += *name++;
	return index & 7;
}

typedef struct {
	FILE* file;
	char* name;
	uint32_t name_length;
	uint32_t size;  /* uncompressed size of file */
	uint32_t zsize; /* compressed size of file */
} dat_entry;

typedef struct {
	FILE* file;
	unsigned int data_size;     /* Total size of the uncompressed files. */
	unsigned int data_zsize;    /* Total size of the compressed files. */
	unsigned int entries;       /* Number of entries. */
	/* TODO: Rename this. */
	dat_entry* dat_entries;     /* The actual entries. */
	unsigned int entries_size;  /* Total size of the file directory. */
	unsigned int entries_zsize; /* Total size of the compressed file directory. */
} dat_archive; /* TODO: Rename this. */

typedef struct {
	uint32_t magic;
	uint32_t size;
	uint32_t zsize;
	uint32_t entries;
} dat_header;

static void
thdat_write_header(dat_archive* archive)
{
	dat_header header;
	unsigned char xheader[sizeof(dat_header)];

	header.magic = 0x31414854;
	header.size = archive->entries_size + 123456789;
	header.zsize = archive->entries_zsize + 987654321;
	header.entries = archive->entries - 0xf7e7f8ac;
	th_encrypt((unsigned char*)&header, xheader, sizeof(dat_header), 0x1B, 0x37, sizeof(dat_header), sizeof(dat_header));

	if (fseek(archive->file, 0, SEEK_SET) == -1) {
		fprintf(stderr, "datpacker: fseek: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	if (fwrite(&xheader, sizeof(dat_header), 1, archive->file) != 1) {
		fprintf(stderr, "datpacker: fwrite: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
}

static void
thdat_write_entries(dat_archive* archive)
{
	unsigned char* buffer = NULL;
	unsigned char* xbuffer = NULL;
	unsigned char* zbuffer = NULL;
	unsigned char* buffer_ptr;
	unsigned int i;
	uint32_t offset = 16;

	buffer = malloc(archive->entries_size);
	if (!buffer) {
		fprintf(stderr, "datpacker: malloc: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	memset(buffer, 0, archive->entries_size);
	buffer_ptr = buffer;

	for (i = 0; i < archive->entries; ++i) {
		dat_entry* entry = &archive->dat_entries[i];
		memcpy(buffer_ptr, entry->name, entry->name_length);
		memcpy(buffer_ptr + entry->name_length, &offset, sizeof(uint32_t));
		memcpy(buffer_ptr + entry->name_length + 4, &entry->size, sizeof(uint32_t));
		buffer_ptr += entry->name_length + 12;
		offset += entry->zsize;
	}

	zbuffer = th_lz(buffer, archive->entries_size, &archive->entries_zsize);
	free(buffer);

	xbuffer = malloc(archive->entries_zsize);
	if (!xbuffer) {
		fprintf(stderr, "datpacker: malloc: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	th_encrypt(zbuffer, xbuffer, archive->entries_zsize, 0x3e, 0x9b, 0x80, archive->entries_size);
	free(zbuffer);

	if (fwrite(xbuffer, archive->entries_zsize, 1, archive->file) != 1) {
		fprintf(stderr, "datpacker: fwrite: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	free(xbuffer);
}

static void
thdat_close(dat_archive* archive)
{
	if (archive->dat_entries) {
		unsigned int i;
		for (i = 0; i < archive->entries; ++i) {
			if (archive->dat_entries[i].name) {
				free(archive->dat_entries[i].name);
				archive->dat_entries[i].name = NULL;
			}
		}
		free(archive->dat_entries);
		archive->dat_entries = NULL;
	}

	if (archive->file) {
		if (fclose(archive->file) == -1) {
			fprintf(stderr, "datpacker: fclose: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		archive->file = NULL;
	}
}

int
main(int argc, char* argv[])
{
	unsigned int i = 0;
	dat_archive archive;

	if (argc < 3) {
		puts("Usage: datpacker ARCHIVE FILE...");
		exit(EXIT_SUCCESS);
	}

	archive.file = NULL;
	archive.data_size = 0;
	archive.data_zsize = 0;
	archive.entries = argc - 2;
	archive.dat_entries = malloc(archive.entries * sizeof(dat_entry));
	if (!archive.dat_entries) {
		fprintf(stderr, "datpacker: malloc: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	archive.entries_size = 0;
	archive.entries_zsize = 0;

	archive.file = fopen(argv[1], "wb");
	if (!archive.file) {
		fprintf(stderr, "datpacker: could not open %s: %s\n", argv[1], strerror(errno));
		exit(EXIT_FAILURE);
	}

	/* Write the header last. */
	if (fseek(archive.file, 16, SEEK_SET) == -1) {
		fprintf(stderr, "datpacker: fseek: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}

	/* Pack the data and prepare the file directory. */
	for (i = 0; i < archive.entries; ++i) {
		dat_entry* entry = &archive.dat_entries[i];
		char* path = strdup(argv[i + 2]);
		const char* name = basename(path);
		unsigned char* data = NULL;
		unsigned char* xdata = NULL;
		unsigned char* zdata = NULL;
		unsigned int ci = 0;
		unsigned int namelen = 0;
		int ret;

		namelen = strlen(name) + 4;
		namelen -= namelen % 4;
		entry->name = malloc(namelen);
		if (!entry->name) {
			fprintf(stderr, "datpacker: malloc: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		memset(entry->name, 0, namelen);
		memcpy(entry->name, name, strlen(name));

		archive.entries_size += namelen + 12;
		entry->name_length = namelen;
		entry->size = 0;
		entry->zsize = 0;
		
		entry->file = fopen(argv[i + 2], "rb");
		if (!entry->file) {
			fprintf(stderr, "datpacker: could not open %s: %s\n", argv[i + 2], strerror(errno));
			exit(EXIT_FAILURE);
		}

		if (fseek(entry->file, 0, SEEK_END) == -1) {
			fprintf(stderr, "datpacker: could not open %s: %s\n", argv[i + 2], strerror(errno));
			exit(EXIT_FAILURE);
		}

		entry->size = ftell(entry->file);
		if (entry->size == -1) {
			fprintf(stderr, "datpacker: ftell: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		archive.data_size += entry->size;

		if (fseek(entry->file, 0, SEEK_SET) == -1) {
			fprintf(stderr, "datpacker: fseek: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}

		data = malloc(entry->size);
		if (!data) {
			fprintf(stderr, "datpacker: malloc: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}

		ret = fread(data, entry->size, 1, entry->file);
		if (ret != 1) {
			fprintf(stderr, "datpacker: fread: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}

		ret = fclose(entry->file);
		if (ret == -1) {
			fprintf(stderr, "datpacker: fclose: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}

		fflush(stdout);
		zdata = th_lz0(data, entry->size, &entry->zsize);
		if (entry->zsize >= entry->size) {
			entry->zsize = entry->size;
			memcpy(zdata, data, entry->size);
		}
		free(data);
		archive.data_zsize += entry->zsize;

		xdata = malloc(entry->zsize);
		if (!xdata) {
			fprintf(stderr, "datpacker: malloc: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}

		ci = th_get_cryprm_index(name);
		free(path);
		th_encrypt(zdata, xdata, entry->zsize, g_cryprm[ci].key, g_cryprm[ci].step, g_cryprm[ci].block, g_cryprm[ci].limit);
		free(zdata);

		if (fwrite(xdata, entry->zsize, 1, archive.file) != 1) {
			fprintf(stderr, "datpacker: fwrite: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		free(xdata);
	}

	thdat_write_entries(&archive);
	thdat_write_header(&archive);
	thdat_close(&archive);

	return EXIT_SUCCESS;
}
